﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Optimization;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Encodings.RealVectorEncoding;

namespace HeuristicLab.Problems.Modifiers {
  [StorableType("39EFBF50-DCC3-4E04-B0C1-C0EF1BC70736")]
  [Item("MultiObjectiveEvaluationCacheProblemModifier", "A problem modifier that caches old evaluations")]
  public class EvaluationCacheProblemModifier : ProblemModifier {
    private readonly object locker = new object();
    [Storable]
    private Dictionary<string, Tuple<double[], IScope>> cache;

    #region Constructors & Cloning
    [StorableConstructor]
    protected EvaluationCacheProblemModifier(StorableConstructorFlag _) : base(_) { }
    protected EvaluationCacheProblemModifier(EvaluationCacheProblemModifier original, Cloner cloner) : base(original, cloner) {
      cache = original?.cache.ToDictionary(x => x.Key, x => Tuple.Create((double[])x.Value.Item1.Clone(), cloner.Clone(x.Value.Item2)));
    }
    public EvaluationCacheProblemModifier() {
      cache = new Dictionary<string, Tuple<double[], IScope>>();
    }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (cache == null) cache = new Dictionary<string, Tuple<double[], IScope>>();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new EvaluationCacheProblemModifier(this, cloner);
    }
    #endregion

    #region ProblemModifier
    public override double[] ModifiedEvaluate(Individual individual, IRandom random) {
      var key = ToKey(individual, Encoding);
      lock (locker) {
        if (cache.TryGetValue(key, out var tuple))
          return tuple.Item1;
      }
      var qualities = base.ModifiedEvaluate(individual, random);
      StoreInCache(key, qualities, individual);
      return qualities;
    }

    public override void Initialize() {
      cache = new Dictionary<string, Tuple<double[], IScope>>();
    }
    #endregion

    #region Helpers
    private void StoreInCache(string key, double[] qualities, Individual individual) {
      IScope individualScope = new Scope();
      individualScope.Variables.AddRange(individual.Values.Select(v => new Variable(v.Key, v.Value)));
      individualScope = (IScope)individualScope.Clone();
      lock (locker) {
        if (!cache.ContainsKey(key))
          cache.Add(key, Tuple.Create(qualities, individualScope));
      }
    }

    private static string ToKey(Individual i, IEncoding enc) {
      switch (enc) {
        case IntegerVectorEncoding _:
          return string.Join(";", i.IntegerVector());
        case RealVectorEncoding _:
          return string.Join(";", i.RealVector());
        default:
          throw new ArgumentException("Encoding not supported");
      }
    }
    #endregion
  }
}
